package net.knarcraft.dynmapcitizens.handler.trait.quests;

import me.pikamug.quests.quests.Quest;
import me.pikamug.quests.quests.components.Planner;
import net.knarcraft.dynmapcitizens.DynmapCitizens;
import net.knarcraft.knarlib.formatting.StringFormatter;
import net.knarcraft.knarlib.formatting.StringReplacer;
import net.knarcraft.knarlib.formatting.TimeFormatter;
import org.jetbrains.annotations.NotNull;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;

import static net.knarcraft.dynmapcitizens.formatting.QuestsTranslatableMessage.QUESTS_PLANNER_COOL_DOWN;
import static net.knarcraft.dynmapcitizens.formatting.QuestsTranslatableMessage.QUESTS_PLANNER_DESCRIPTION;
import static net.knarcraft.dynmapcitizens.formatting.QuestsTranslatableMessage.QUESTS_PLANNER_FROM;
import static net.knarcraft.dynmapcitizens.formatting.QuestsTranslatableMessage.QUESTS_PLANNER_REPEAT;
import static net.knarcraft.dynmapcitizens.formatting.QuestsTranslatableMessage.QUESTS_PLANNER_TIMESTAMP_FORMAT;
import static net.knarcraft.dynmapcitizens.formatting.QuestsTranslatableMessage.QUESTS_PLANNER_UNREPEATABLE;
import static net.knarcraft.dynmapcitizens.formatting.QuestsTranslatableMessage.QUESTS_PLANNER_UNTIL;

/**
 * A class to generate a string containing all information about a quest's planner info
 */
public class QuestPlannerInfoGenerator {

    private final Quest quest;
    private final StringFormatter formatter;

    /**
     * Instantiates a new quest planner info generator
     *
     * @param quest <p>The quest to generate information about</p>
     */
    public QuestPlannerInfoGenerator(@NotNull Quest quest) {
        this.quest = quest;
        this.formatter = DynmapCitizens.getFormatter();
    }

    /**
     * Gets information about the time-availability of a quest
     *
     * @return <p>Information about when the quest is available</p>
     */
    public @NotNull String getQuestPlannerInfo() {
        Planner planner = quest.getPlanner();
        StringReplacer replacer = new StringReplacer(formatter.getUnFormattedColoredMessage(QUESTS_PLANNER_DESCRIPTION));

        //Quest can be repeated after a cool-down
        if (planner.hasCooldown()) {
            replacer.add("{questCoolDown}", formatter.replacePlaceholder(QUESTS_PLANNER_COOL_DOWN, "{coolDown}",
                    TimeFormatter.getDurationString(DynmapCitizens.getTranslator(), planner.getCooldown() / 1000)));
        } else {
            replacer.add("{questCoolDown}", formatter.getUnFormattedColoredMessage(QUESTS_PLANNER_UNREPEATABLE));
        }

        //Quest only becomes available after the start date
        replacer.add("{questFrom}", planner.hasStart() ? formatter.replacePlaceholder(QUESTS_PLANNER_FROM,
                "{startDate}", formatTimestamp(planner.getStartInMillis())) : "");

        //Quest is only available until the end date
        replacer.add("{questUntil}", planner.hasEnd() ? formatter.replacePlaceholder(QUESTS_PLANNER_UNTIL,
                "{endDate}", formatTimestamp(planner.getEndInMillis())) : "");

        //Quest availability repeats'
        replacer.add("{questRepeat}", planner.hasRepeat() ? formatter.replacePlaceholder(QUESTS_PLANNER_REPEAT,
                "{repeatDelay}", TimeFormatter.getDurationString(DynmapCitizens.getTranslator(),
                        planner.getRepeat() / 1000)) : "");

        return replacer.replace();
    }

    /**
     * Gets a datetime string for the given timestamp
     *
     * @param timestamp <p>A timestamp in milliseconds</p>
     * @return <p>A datetime string</p>
     */
    private @NotNull String formatTimestamp(long timestamp) {
        DateFormat format = new SimpleDateFormat(formatter.getUnFormattedMessage(QUESTS_PLANNER_TIMESTAMP_FORMAT));
        Date date = new Date(timestamp);
        return format.format(date);
    }

}
