package net.knarcraft.dynmapcitizens.handler.trait.quests;

import me.pikamug.quests.Quests;
import me.pikamug.quests.quests.Quest;
import me.pikamug.quests.quests.components.Stage;
import net.citizensnpcs.api.CitizensAPI;
import net.citizensnpcs.api.npc.NPCRegistry;
import net.knarcraft.dynmapcitizens.DynmapCitizens;
import net.knarcraft.dynmapcitizens.formatting.QuestsTranslatableMessage;
import net.knarcraft.dynmapcitizens.handler.trait.AbstractTraitHandler;
import net.knarcraft.dynmapcitizens.property.Icon;
import net.knarcraft.dynmapcitizens.settings.QuestsSettings;
import net.knarcraft.dynmapcitizens.settings.TraitSettings;
import net.knarcraft.dynmapcitizens.util.QuestsHelper;
import net.knarcraft.knarlib.formatting.StringFormatter;
import net.knarcraft.knarlib.formatting.StringReplacer;
import org.bukkit.Bukkit;
import org.dynmap.DynmapAPI;
import org.dynmap.markers.GenericMarker;
import org.dynmap.markers.Marker;
import org.dynmap.markers.MarkerIcon;
import org.jetbrains.annotations.NotNull;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import static net.knarcraft.dynmapcitizens.formatting.QuestsTranslatableMessage.QUESTS_INVOLVED_IN_DELIVERY_TARGET;
import static net.knarcraft.dynmapcitizens.formatting.QuestsTranslatableMessage.QUESTS_INVOLVED_IN_FORMAT;
import static net.knarcraft.dynmapcitizens.formatting.QuestsTranslatableMessage.QUESTS_INVOLVED_IN_INTERACT;
import static net.knarcraft.dynmapcitizens.formatting.QuestsTranslatableMessage.QUESTS_INVOLVED_IN_KILL;
import static net.knarcraft.dynmapcitizens.formatting.QuestsTranslatableMessage.QUESTS_MARKER_NAME_FORMAT;
import static net.knarcraft.dynmapcitizens.formatting.QuestsTranslatableMessage.QUESTS_OFFERED_FORMAT;
import static net.knarcraft.dynmapcitizens.formatting.QuestsTranslatableMessage.QUESTS_OFFERED_QUEST_FORMAT;

/**
 * A handler class for the quests trait
 */
public class QuestsHandler extends AbstractTraitHandler {

    private Quests questsAPI;
    private Map<Icon, MarkerIcon> markerIcons;
    private final Map<UUID, NPCQuestInfo> questGiverInfo = new HashMap<>();
    private final QuestsSettings settings = new QuestsSettings();
    private final List<Quest> unavailableQuests = new ArrayList<>();
    private QuestAreaHandler questAreaHandler;

    @Override
    public void initialize() {
        questsAPI = (Quests) Bukkit.getServer().getPluginManager().getPlugin("Quests");
        DynmapAPI dynmapAPI = DynmapCitizens.getInstance().getDynmapAPI();
        markerIcons = DynmapCitizens.getInstance().getGlobalSettings().getMarkerIcons();
        if (questsAPI != null) {
            super.initializeMarkerSet();
            super.markerSet.getMarkers().forEach(GenericMarker::deleteMarker);

            questAreaHandler = new QuestAreaHandler(questsAPI, dynmapAPI, settings, unavailableQuests);
        } else {
            isEnabled = false;
        }
    }

    @Override
    public @NotNull TraitSettings getSettings() {
        return this.settings;
    }

    @Override
    public void updateMarkers() {
        if (questsAPI.isLoading()) {
            return;
        }

        //Generate information about all NPCs involved in quests
        generateQuestNPCInfo();

        //Update all area markers for quests
        questAreaHandler.updateQuestAreas();

        //Remove any markers for deleted quests
        for (Marker marker : super.markerSet.getMarkers()) {
            if (!questGiverInfo.containsKey(UUID.fromString(marker.getMarkerID()))) {
                marker.deleteMarker();
            }
        }

        //Generate markers based on the generated info
        generateAllMarkers();
    }

    /**
     * Generates information about all NPCs involved in quests
     */
    private void generateQuestNPCInfo() {
        //Clear any previous information
        questGiverInfo.clear();
        unavailableQuests.clear();
        //Generation information about NPC's parts in each quest
        for (Quest quest : questsAPI.getLoadedQuests()) {
            if (QuestsHelper.isQuestUnavailable(quest)) {
                unavailableQuests.add(quest);
                //If unavailable quests aren't displayed, there is no point in continuing
                if (settings.hideUnavailableQuests()) {
                    continue;
                }
            }
            if (quest.getNpcStart() != null) {
                getInfo(quest.getNpcStart()).addQuestStart(quest);
            }
            for (Stage stage : quest.getStages()) {
                for (UUID npcId : stage.getNpcsToKill()) {
                    getInfo(npcId).addQuestKill(quest);
                }
                for (UUID npcId : stage.getItemDeliveryTargets()) {
                    getInfo(npcId).addQuestDeliver(quest);
                }
                for (UUID npcId : stage.getNpcsToInteract()) {
                    getInfo(npcId).addQuestInteract(quest);
                }
            }
        }
    }

    /**
     * Generates all quest markers based on the previously generated quest NPC info
     */
    private void generateAllMarkers() {
        NPCRegistry registry = CitizensAPI.getNPCRegistry();
        //Add markers for each NPC detected as part of a quest
        for (UUID npcId : questGiverInfo.keySet()) {
            NPCQuestInfo info = questGiverInfo.get(npcId);
            MarkerIcon icon = markerIcons.get(info.getNPCIcon());
            List<Quest> offeredQuests = info.getQuestStarts();
            StringBuilder markerDescription = new StringBuilder();

            markerDescription.append(DynmapCitizens.getFormatter().replacePlaceholder(QUESTS_MARKER_NAME_FORMAT,
                    "{name}", registry.getByUniqueId(npcId).getName()));

            //Print info about any offered quests
            if (!offeredQuests.isEmpty()) {
                appendOfferedQuestsInfo(markerDescription, offeredQuests);
            }

            markerDescription.append(getInvolvedInQuestsString(info));

            Marker existingMarker = super.markerSet.findMarker(npcId.toString());
            String newDescription = markerDescription.toString();
            if (existingMarker != null) {
                if (!existingMarker.getDescription().equals(newDescription)) {
                    existingMarker.setDescription(newDescription);
                }
            } else {
                addNPCMarker(npcId, info.getQuestNPCType().getMarkerTitle(), newDescription, icon, super.markerSet);
            }
        }
    }

    /**
     * Appends information about an NPC's offered quests to the given string builder
     *
     * @param stringBuilder <p>The string builder to append to</p>
     * @param offeredQuests <p>The list of quests the NPC offers</p>
     */
    private void appendOfferedQuestsInfo(@NotNull StringBuilder stringBuilder, @NotNull List<Quest> offeredQuests) {
        StringFormatter formatter = DynmapCitizens.getFormatter();
        StringBuilder questsInfoBuilder = new StringBuilder();

        for (Quest quest : offeredQuests) {
            StringReplacer replacer = new StringReplacer(formatter.getUnFormattedMessage(QUESTS_OFFERED_QUEST_FORMAT));
            replacer.add("{name}", quest.getName());
            replacer.add("{description}", quest.getDescription());

            replacer.add("{stages}", settings.displayStageInfo() ?
                    new QuestStagesInfoGenerator(quest).getQuestStagesInfo() : "");
            replacer.add("{rewards}", settings.displayRewardInfo() ?
                    new QuestRewardsInfoGenerator(quest).getQuestRewardsInfo() : "");
            replacer.add("{requirements}", settings.displayRequirementInfo() ?
                    new QuestRequirementsInfoGenerator(questsAPI, quest).getQuestRequirementsInfo() : "");
            replacer.add("{planner}", settings.displayPlannerInfo() ?
                    new QuestPlannerInfoGenerator(quest).getQuestPlannerInfo() : "");
            questsInfoBuilder.append(replacer.replace());
        }

        stringBuilder.append(formatter.replacePlaceholder(QUESTS_OFFERED_FORMAT, "{quests}",
                questsInfoBuilder.toString()));
    }

    /**
     * Gets information about the quests an NPC is involved in
     *
     * @param info <p>The NPC info to look through</p>
     * @return <p>Information about an NPC's involvement in different quests</p>
     */
    private String getInvolvedInQuestsString(@NotNull NPCQuestInfo info) {
        StringFormatter formatter = DynmapCitizens.getFormatter();
        List<Quest> questKills = info.getQuestKills();
        List<Quest> questInteractions = info.getQuestInteractions();
        List<Quest> questDeliveries = info.getQuestDeliveries();

        if (questKills.isEmpty() && questInteractions.isEmpty() && questDeliveries.isEmpty()) {
            return "";
        }
        StringBuilder markerDescription = new StringBuilder();

        addInvolvedInString(formatter, QUESTS_INVOLVED_IN_KILL, questKills, markerDescription);
        addInvolvedInString(formatter, QUESTS_INVOLVED_IN_DELIVERY_TARGET, questDeliveries, markerDescription);
        addInvolvedInString(formatter, QUESTS_INVOLVED_IN_INTERACT, questInteractions, markerDescription);

        return formatter.replacePlaceholder(QUESTS_INVOLVED_IN_FORMAT, "{involvedQuests}",
                markerDescription.toString());
    }

    /**
     * Adds string explaining how an NPC is involved with the given quests
     *
     * @param formatter           <p>The string formatter to use for formatting</p>
     * @param translatableMessage <p>The translatable message to display</p>
     * @param quests              <p>The quests the NPC is involved in</p>
     * @param builder             <p>The string builder to append to</p>
     */
    private void addInvolvedInString(@NotNull StringFormatter formatter,
                                     @NotNull QuestsTranslatableMessage translatableMessage, @NotNull List<Quest> quests,
                                     @NotNull StringBuilder builder) {
        for (Quest quest : new HashSet<>(quests)) {
            builder.append(formatter.replacePlaceholder(translatableMessage, "{questName}", quest.getName()));
        }
    }


    /**
     * Gets the info object for the given NPC
     *
     * @param npcId <p>The id of the NPC to get information about</p>
     * @return <p>The NPC's info object</p>
     */
    private @NotNull NPCQuestInfo getInfo(@NotNull UUID npcId) {
        if (questGiverInfo.get(npcId) == null) {
            questGiverInfo.put(npcId, new NPCQuestInfo());
        }
        return questGiverInfo.get(npcId);
    }

}
